/***************************************************************************/
/*                                                                         */
/*                             DLLIST.C                                    */
/*                                                                         */
/*                     Generic doubly linked list                          */
/*                                                                         */
/*                                                                         */
/***************************************************************************/



#include "dllist.h"


/*
   This speeds up quicksort by bypassing the need to pass the function
   pointer to the comparison function in the quick sort parameter list
   every time it is called. Cannot be used in a threaded application.
*/

static CompareFunction_t * storedCompare;



/************************* private function prototypes *********************/


static void List_insertionSortLink ( ListNode_t ** listHead,
                                      ListNode_t * listEnd,
                                       CompareFunction_t Compare );

static void List_repairPriorLinks ( List_t * const pList );

static void List_xQuickSortL2 ( ListNode_t **Head, ListNode_t * End, long N );



/***************************************************************************/
/*                                                                         */
/*                         Constructor/Destructor                          */
/*                                                                         */
/***************************************************************************/



extern List_t * List_new ( CompareFunction_t * pfnCompare,
                            CreateObjectFunction_t * pfnCreateData,
                             FreeObjectFunction_t * pfnFreeData )
/*
 TASK : Creates a new list.
 IN   : Pointers to the user supplied comparison, memory allocation and
        memory deallocation functions.
 OUT  : Pointer to the newly created list. If NULL is returned, the list
        could not be created.
*/
{
   List_t *pList = NULL;

   if ( NULL == (pList = (List_t *) malloc (sizeof (List_t))))
      return NULL;

   /* set up dummy tail node */
   pList->finalNode.pNextNode =
      pList->finalNode.pPriorNode = &(pList->finalNode);
   pList->pHeadNode = pList->pCurrentNode = &(pList->finalNode);
   pList->finalNode.pvNodeData = NULL;
   pList->numberOfNodes = pList->numberOfCurrentNode = 0;
   pList->walkingListStatus = currentlyNotWalkingList;
   pList->pfnCompare = pfnCompare;
   pList->pfnCreateData = pfnCreateData;
   pList->pfnFreeData = pfnFreeData;
   return pList;
}


extern void List_delete (List_t **ppList)
/* TASK : Deletes the list */
{
   ListNode_t * pFirstNode;
   ListNode_t * pRestNode;
   List_t *pList = *ppList;

   if ( pList == NULL )
      return;
   if ( pList->pHeadNode != &(pList->finalNode))
   {
      pFirstNode = pList->pHeadNode;
      pRestNode = pList->pHeadNode->pNextNode;
      while (pFirstNode != &(pList->finalNode))
      {
         pList->pfnFreeData (pFirstNode->pvNodeData);
         free (pFirstNode);
         pFirstNode = pRestNode;
         pRestNode = pRestNode->pNextNode;
      }
      pList->numberOfNodes = pList->numberOfCurrentNode = 0;
      pList->pHeadNode = pList->pCurrentNode = &(pList->finalNode);
   }/* end if not empty list */
   if ( pList != NULL )
      free (pList);
   *ppList = NULL;
}




/***************************************************************************/
/*                                                                         */
/*                            Enquiry Operators                            */
/*                                                                         */
/***************************************************************************/


extern Boolean List_isEmpty (List_t * const pList)
/*
 TASK : Returns Btrue if the list pointed to by pList contains no nodes.
*/
{
   assert (pList != NULL);

   if (pList->pHeadNode == &(pList->finalNode))
     return Btrue;
   return Bfalse;
}


extern long List_currentPosition (List_t * const pList)
/*
 TASK : Returns the position in the list of the current node ( Head node
        ordinal position is 1 )
*/
{
   assert (pList != NULL);
   return pList->numberOfCurrentNode;
}


extern Boolean List_notAtEnd (List_t * const pList)
/*
 TASK : Returns Btrue if the current node is not the tail node.
*/
{
   assert (pList != NULL);
   if (pList->pCurrentNode->pNextNode ==  &(pList->finalNode))
     return Bfalse;
   return Btrue;
}


extern long List_numberOfNodes (List_t * const pList)
/*
  TASK : Returns the number of nodes in the list
*/
{
   assert (pList != NULL);
   return pList->numberOfNodes;
}




/***************************************************************************/
/*                                                                         */
/*                         Positional Operators                            */
/*                                                                         */
/***************************************************************************/



extern Boolean List_moveToHead (List_t * const pList)
/*
 TASK : Makes the head node the current node. If the list is empty, returns
        Bfalse, otherwise returns Btrue.
 NOTE : Incompatable with sequential traversal.
*/
{
   assert (pList != NULL);
   if (pList->pHeadNode == &(pList->finalNode))
      return Bfalse;
   pList->pCurrentNode = pList->pHeadNode;
   pList->numberOfCurrentNode = 1;
   pList->walkingListStatus = currentlyNotWalkingList;
   return Btrue;
}



extern Boolean List_moveToTail (List_t * const pList)
/*
 TASK : Makes the tail node the current node
 NOTE : Incompatable with sequential traversal.
*/
{
   assert (pList != NULL);
   if (pList->pHeadNode == &(pList->finalNode))
      return Bfalse;
   pList->pCurrentNode = pList->finalNode.pPriorNode;
   pList->numberOfCurrentNode = pList->numberOfNodes;
   pList->walkingListStatus = currentlyNotWalkingList;
   return Btrue;
}/* List_moveToTail */



extern Boolean List_moveNthNode (List_t * const pList, long MoveToThisNode)
/*
 TASK : Makes the nth node of the list the current node. Returns Bfalse if the
        list is empty, or MoveToThisNode is greater than the number of nodes
        in the list, otherwise returns Btrue.
 NOTE : Incompatable with sequential traversal. The head node position is 1.
*/
{
   long count;

   assert (pList != NULL);
   if ((MoveToThisNode > pList->numberOfNodes) || (MoveToThisNode < 1) || pList->pHeadNode == &(pList->finalNode))
      return Bfalse;
   if ( MoveToThisNode == pList->numberOfCurrentNode )
      return Btrue;
   pList->pCurrentNode = pList->pHeadNode;
   for (count = 1; count < MoveToThisNode; ++count)
      pList->pCurrentNode = pList->pCurrentNode->pNextNode;
   pList->numberOfCurrentNode = MoveToThisNode;
   pList->walkingListStatus = currentlyNotWalkingList;
   return Btrue;
}



extern Boolean List_moveToNextNode (List_t * const pList)
/*
 TASK : Makes the next node of the list the current node. If the tail of
        the list is reached moves to the list head and returns Bfalse,
        otherwise returns Btrue.
 NOTE : Incompatable with sequential traversal.
*/
{
   assert (pList != NULL);
   if (pList->pHeadNode != &(pList->finalNode))
   {
      pList->walkingListStatus = currentlyNotWalkingList;
      pList->pCurrentNode = pList->pCurrentNode->pNextNode;
      if (pList->pCurrentNode != &(pList->finalNode))
      {
         if (pList->pCurrentNode != pList->pHeadNode)
            ++(pList->numberOfCurrentNode);
         else
            pList->numberOfCurrentNode = 1;
         return Btrue;
      }
      else
      { /* reached past the end of the list so move to head */
         pList->pCurrentNode = pList->pHeadNode;
         pList->numberOfCurrentNode = 1;
         return Bfalse;
      }
   }/* end if not empty list */
   return Bfalse;
}


extern Boolean List_moveToPreviousNode (List_t * const pList)
/*
 TASK : Makes the previous node in the list the current node. If the head of
        the list the list is reached moves to the list tail and returns Bfalse
        otherwise returns Btrue.
 NOTE : Incompatable with sequential traversal.
*/
{
   assert (pList != NULL);
   if (pList->pHeadNode != &(pList->finalNode))
   {
      pList->walkingListStatus = currentlyNotWalkingList;
      pList->pCurrentNode = pList->pCurrentNode->pPriorNode;
      if (pList->pCurrentNode != &(pList->finalNode))
      {
         if (pList->pCurrentNode != pList->finalNode.pPriorNode)
            --(pList->numberOfCurrentNode);
         else
            pList->numberOfCurrentNode = pList->numberOfNodes;
         return Btrue;
      }
      else
      { /* reached past beginning of the list so move to tail */
         pList->pCurrentNode = pList->finalNode.pPriorNode;
         pList->numberOfCurrentNode = pList->numberOfNodes;
         return Bfalse;
      }
   }/* end if not empty list */
   return Bfalse;
}


extern Boolean List_makeCurrentHead ( List_t * const pList )
/*
 TASK : Makes the current node the head of the list.
        If the list is empty then returns Bfalse, otherwise returns Btrue
*/
{
   assert (pList != NULL);
   if ( pList->pHeadNode != &(pList->finalNode) && pList->pCurrentNode != &(pList->finalNode))
   {
      pList->finalNode.pPriorNode->pNextNode = pList->pHeadNode;
      pList->pHeadNode->pPriorNode = pList->finalNode.pPriorNode;
      pList->pCurrentNode->pPriorNode->pNextNode = &(pList->finalNode);
      pList->finalNode.pPriorNode = pList->pCurrentNode->pPriorNode;
      pList->pCurrentNode->pPriorNode = &(pList->finalNode);
      pList->finalNode.pNextNode = pList->pCurrentNode;
      pList->pHeadNode = pList->pCurrentNode;
      pList->numberOfCurrentNode = 1;
      pList->walkingListStatus = currentlyNotWalkingList;
      return Btrue;
   }
   return Bfalse;
}


/***************************************************************************/
/*                                                                         */
/*                         Extraction Operators                            */
/*                                                                         */
/***************************************************************************/


extern void * List_data ( List_t * const pList )
/*
 TASK : Returns a pointer to the contents of the current node.
*/
{
   assert (pList != NULL);
   if (pList->pHeadNode == &(pList->finalNode))
      return NULL;
   return pList->pCurrentNode->pvNodeData;
}


extern void * const List_next ( List_t * const pList )
/*
 TASK : Returns a pointer to the next node in the list or NULL if there
        is no next node
*/
{
   assert (pList != NULL);
   if ( pList->pHeadNode == &(pList->finalNode))
      return NULL;
   if ( pList->pCurrentNode->pNextNode != &(pList->finalNode))
   {
      return pList->pCurrentNode->pNextNode->pvNodeData;
   }
   return NULL;
}


extern ListNode_t * List_nodePointer ( List_t * const pList )
/*
 TASK : Returns a pointer to the current node.
*/
{
   assert (pList != NULL);
   if (pList->pHeadNode == &(pList->finalNode))
      return NULL;
   return pList->pCurrentNode;
}


extern void * const List_walk ( List_t * const pList )
/*
 TASK : Sequentially returns a pointer to the contents of each node in the
        list starting at the head node and moving to the tail.
        Returns NULL if the list is empty or if reached the end of list.
        Moves to head node when traversal complete.

        Always call List_moveToHead before List_walk if you need to be
        absolutely certain the list walk will start from the first node

 NOTE : IT IS CRITICALLY IMPORTANT THAT IF YOU TERMINATE WALKING THE LIST
        BEFORE THE NULL POINTER IS RETURNED YOU CALL List_moveToHead TO
        RESET THE LIST STATE.
*/
{
   assert (pList != NULL);
   if ( pList->pHeadNode == &(pList->finalNode))
      return NULL;
   if ( pList->pCurrentNode == &(pList->finalNode))
      pList->walkingListStatus = currentlyNotWalkingList;
   if ( pList->walkingListStatus == currentlyNotWalkingList )
   {
      pList->pCurrentNode = &(pList->finalNode);
      pList->numberOfCurrentNode = 0;
      pList->walkingListStatus = currentlyWalkingList;
   }
   if ( pList->pCurrentNode->pNextNode != &(pList->finalNode))
   {
      pList->pCurrentNode = pList->pCurrentNode->pNextNode;
      ++(pList->numberOfCurrentNode);
   }
   else
   {
      pList->walkingListStatus = currentlyNotWalkingList;
      pList->pCurrentNode = pList->pHeadNode;
      pList->numberOfCurrentNode = 1;
      return NULL;
   }
   return pList->pCurrentNode->pvNodeData;
}


extern void * const List_walkFromNth ( List_t * const pList, long NodeNumber )
/*
 TASK : Sequentially returns a pointer to the contents of each node in the
        list starting at the NodeNumber node and moving to the tail.
        Returns NULL if the list is empty or if reached the end of list,
        or if NodeNumber is > number of nodes in the list
        Moves to head node when traversal complete.

        Always call List_moveToHead before List_walkFromNth if you need
        to be absolutely certain the walk will start from the first node

 NOTE : IT IS CRITICALLY IMPORTANT THAT IF YOU TERMINATE WALKING THE LIST
        BEFORE THE NULL POINTER IS RETURNED YOU CALL List_moveToHead TO
        RESET THE LIST STATE.
*/
{
   assert (pList != NULL);
   if ( pList->pHeadNode == &(pList->finalNode))
      return NULL;
   if ( pList->pCurrentNode == &(pList->finalNode))
      pList->walkingListStatus = currentlyNotWalkingList;
   if ( pList->walkingListStatus == currentlyNotWalkingList )
   {
      if ( ! List_moveNthNode ( pList, NodeNumber - 1))
         pList->walkingListStatus = currentlyWalkingList;
   }

   if ( pList->pCurrentNode->pNextNode != &(pList->finalNode))
   {
      pList->pCurrentNode = pList->pCurrentNode->pNextNode;
      ++(pList->numberOfCurrentNode);
   }
   else
   {
      pList->walkingListStatus = currentlyNotWalkingList;
      pList->pCurrentNode = pList->pHeadNode;
      pList->numberOfCurrentNode = 1;
      return NULL;
   }
   return pList->pCurrentNode->pvNodeData;
}


extern void * const List_walkAndFind ( List_t * const pList, const void * pvData )
/*
 TASK : Starting at the head node and traversing towards the tail of the list
        this function returns a pointer to the next instance of a node found
        to be identical to the data referenced by pvData according to the
        comparison function pfnCompare. Returns NULL if no more hits are found
        or the list is empty.

        Always call List_moveToHead before List_walkAndFind if you need
        to be absolutely certain the search will start from the first node

 NOTE : IT IS CRITICALLY IMPORTANT THAT IF YOU TERMINATE WALKING THE LIST
        BEFORE THE NULL POINTER IS RETURNED YOU CALL List_moveToHead TO
        RESET THE LIST STATE.
*/
{
   assert (pList != NULL);
   if ( pList->pHeadNode == &(pList->finalNode))
      return NULL;
   if ( pList->pCurrentNode == &(pList->finalNode))
      pList->walkingListStatus = currentlyNotWalkingList;
   if ( pList->walkingListStatus == currentlyNotWalkingList )
   {
      pList->pCurrentNode = &(pList->finalNode);
      pList->numberOfCurrentNode = 0;
      pList->walkingListStatus = currentlyWalkingList;
   }
   do
   {
      if ( pList->pCurrentNode->pNextNode != &(pList->finalNode))
      { /* not at end of the list */
         pList->pCurrentNode = pList->pCurrentNode->pNextNode;
         ++(pList->numberOfCurrentNode);
         if (0 == pList->pfnCompare (pList->pCurrentNode->pvNodeData, pvData))
            return pList->pCurrentNode->pvNodeData;
      }
      else
      {
         pList->pCurrentNode = pList->pHeadNode;
         pList->numberOfCurrentNode = 1;
         pList->walkingListStatus = currentlyNotWalkingList;
         return NULL;
      }
   }while (Btrue);
}


extern void * const List_firstMatch (List_t * const pList, const void * pvData )
/*
 TASK : Assuming the list is walked from the first node to the last, this
        will return a pointer to the first node whose data is found to
        match pvData and also make that node current. Returns NULL if
        no match was found.
*/
{
   ListNode_t *pTempNode = NULL;
   long count = 1;

   assert ( pList != NULL );
   if ( pList->pHeadNode == &(pList->finalNode ))
      return NULL;
   for ( pTempNode = pList->pHeadNode; pTempNode != &(pList->finalNode); pTempNode = pTempNode->pNextNode, ++count )
   {
      if (0 == pList->pfnCompare ( pTempNode->pvNodeData, pvData ))
      {
         pList->pCurrentNode = pTempNode;
         pList->numberOfCurrentNode = count;
         pList->walkingListStatus = currentlyNotWalkingList;;
         return pList->pCurrentNode->pvNodeData;
      }
   }
   pList->walkingListStatus = currentlyNotWalkingList;;
   return NULL;
}




/***************************************************************************/
/*                                                                         */
/*                          Inclusion Operators                            */
/*                                                                         */
/***************************************************************************/


extern void List_replaceCurrent (List_t * const pList, const void * pvData )
/*
 TASK : Replaces the contents of the current node with a copy of the data
        referenced via pvData
*/
{
   assert (pList != NULL);
   assert (pvData != NULL);

   if ( pList->pHeadNode != &(pList->finalNode) &&
         pList->pCurrentNode != &(pList->finalNode ))
   {
     pList->pfnFreeData (pList->pCurrentNode->pvNodeData);
     pList->pCurrentNode->pvNodeData = pList->pfnCreateData ( pvData );
   }
}


extern Boolean List_addAfter ( List_t * const pList, const void * pvData)
/*
 TASK : Adds a new node to the list containing a copy of the data referenced
        by pvData after the current node. Returns Btrue if memory was
        avaliable to create the new node, otherwise returns Bfalse.
*/
{
   ListNode_t *pTempNode = NULL;

   assert (pList != NULL);
   assert (pvData != NULL);

   if (NULL == (pTempNode = malloc (sizeof (ListNode_t))))
      return Bfalse;
   if (NULL == (pTempNode->pvNodeData = pList->pfnCreateData ( pvData )))
      return Bfalse;
   pList->pCurrentNode->pNextNode->pPriorNode = pTempNode;
   pTempNode->pNextNode = pList->pCurrentNode->pNextNode;
   pTempNode->pPriorNode = pList->pCurrentNode;
   pList->pCurrentNode->pNextNode = pTempNode;
   ++(pList->numberOfCurrentNode);
   ++(pList->numberOfNodes);
   if ( pList->pCurrentNode == &(pList->finalNode))
   { /* The list was empty */
      pList->pHeadNode = pTempNode;
   }
   pList->pCurrentNode = pTempNode;
   pList->walkingListStatus = currentlyNotWalkingList;
   return Btrue;
}


extern Boolean List_addBefore ( List_t * const pList, const void * pvData )
/*
 TASK : Adds a new node to the list containing a copy of the data referenced
        by pvData before the current node. Returns Btrue if memory was
        avaliable to create the new node, otherwise returns Bfalse.
*/
{
   assert (pList != NULL);
   assert ( pvData != NULL );

   if (pList->pHeadNode != &(pList->finalNode))
   {
      pList->pCurrentNode = pList->pCurrentNode->pPriorNode;
      --(pList->numberOfCurrentNode);
      if (! List_addAfter (pList,pvData))
         return Bfalse;
   }
   else
   { /* At the head of the list, so insert before head and make this the
        new head node */
      pList->pCurrentNode = &(pList->finalNode);
      if (! List_addAfter (pList,pvData))
         return Bfalse;
      pList->pHeadNode = pList->pCurrentNode;
   }
   return Btrue;
}


extern Boolean List_addList ( List_t * const pListA, List_t * const pListB)
/*
 TASK : Adds the contents of listB after the current node of listA.
        Returns Btrue if memory was avaliable for the operation,
        otherwise returns Bfalse.
*/
{
   ListNode_t *pTempNode = NULL;

   assert ( pListA != NULL );
   assert ( pListB != NULL );

   if ( pListB->pHeadNode == &(pListB->finalNode ))
     return Bfalse;           /* no point appending if nothing to append */
   for ( pTempNode = pListB->pHeadNode; pTempNode != &(pListB->finalNode); pTempNode = pTempNode->pNextNode )
   {
      if ( !List_addAfter ( pListA, pTempNode->pvNodeData))
         return Bfalse;
   }
   pListA->walkingListStatus = currentlyNotWalkingList;
   pListB->walkingListStatus = currentlyNotWalkingList;
   return Btrue;
}


extern Boolean List_appendList  ( List_t * const pListA, List_t * const pListB )
/*
 TASK : Appends a copy of pListB from the current node in ListB to the
        end of ListB to the end of pListA. The list referenced
        by pListB is unaffected by the append. Returns Btrue if memory was
        avaliable to make the copy of pListB, otherwise returns Bfalse.
*/
{
   ListNode_t *pTempNode = NULL;

   assert ( pListA != NULL );
   assert ( pListB != NULL );

   if ( pListB->pHeadNode == &(pListB->finalNode ))
      return Bfalse;           /* no point appending if nothing to append */
   pListA->pCurrentNode = pListA->finalNode.pPriorNode;
   pListA->numberOfCurrentNode = pListA->numberOfNodes;
   for ( pTempNode = pListB->pHeadNode; pTempNode != &(pListB->finalNode); pTempNode = pTempNode->pNextNode )
   {
      if ( !List_addAfter ( pListA, pTempNode->pvNodeData))
         return Bfalse;
   }
   pListA->walkingListStatus = currentlyNotWalkingList;
   pListB->walkingListStatus = currentlyNotWalkingList;
   return Btrue;
}




/***************************************************************************/
/*                                                                         */
/*                           Deletion Operators                            */
/*                                                                         */
/***************************************************************************/


extern Boolean List_removeCurrent (List_t * const pList)
/*
 TASK : Deletes the current node from the list. Returns Bfalse if the list is
        empty, otherwise returns Btrue.
*/
{
   ListNode_t *pPreCurrent = NULL;

   assert (pList != NULL);
   if ( pList->pHeadNode != &(pList->finalNode) && pList->pCurrentNode != &(pList->finalNode))
   {
      pPreCurrent = pList->pCurrentNode;
      pList->pCurrentNode->pPriorNode->pNextNode = pList->pCurrentNode->pNextNode;
      pList->pCurrentNode->pNextNode->pPriorNode = pList->pCurrentNode->pPriorNode;
      if (pList->pCurrentNode == pList->pHeadNode)
      /* node being removed is the head node */
         pList->pHeadNode = pList->pCurrentNode->pNextNode;
      if ( pList->pCurrentNode->pNextNode != &(pList->finalNode))
      {
         pList->pCurrentNode = pList->pCurrentNode->pNextNode;
      }
      else
      {
         pList->pCurrentNode = pList->pCurrentNode->pPriorNode;
         --(pList->numberOfCurrentNode);
      }
      --(pList->numberOfNodes);
      pList->pfnFreeData(pPreCurrent->pvNodeData);
      pPreCurrent->pvNodeData = NULL;
      free (pPreCurrent);
      pPreCurrent = NULL;
      pList->walkingListStatus = currentlyNotWalkingList;;
      return Btrue;
   }/* end if not empty list*/
   else
      return Bfalse;
}


extern void List_empty ( List_t * const pList )
/*
 TASK : Removes all the nodes in the list pointed to by pList
*/
{
   ListNode_t * pFirstNode;
   ListNode_t * pRestNode;

   assert (pList != NULL);
   if ( pList->pHeadNode != &(pList->finalNode))
   {
      pFirstNode = pList->pHeadNode;
      pRestNode = pList->pHeadNode->pNextNode;
      while (pFirstNode != &(pList->finalNode))
      {
         pList->pfnFreeData (pFirstNode->pvNodeData);
         free (pFirstNode);
         pRestNode = pRestNode->pNextNode;
      }
      pList->numberOfNodes = pList->numberOfCurrentNode = 0;
      pList->finalNode.pNextNode = pList->finalNode.pPriorNode = pList->pHeadNode = pList->pCurrentNode = &(pList->finalNode);
      pList->walkingListStatus = currentlyNotWalkingList;;
   }/* end if not empty list */
}




/***************************************************************************/
/*                                                                         */
/*                           Sorting Operators                             */
/*                                                                         */
/***************************************************************************/


extern void List_quickSort ( List_t * const pList )
/*
 TASK : Performs a sort on pList using the quicksort algorithm.
 NOTE : Eliminates tail recursion to minimise stack size.
        Uses a randomly selected pivot to avoid problems with pre-ordered
        data.
        Adapted from artice in "C Gazette", Vol 5, No. 6 (1995) by Jeff Taylor
*/
{
   assert (pList != NULL);
   storedCompare = pList->pfnCompare;
   List_xQuickSortL2 ( &(pList->pHeadNode), &(pList->finalNode), pList->numberOfNodes );
   List_repairPriorLinks ( pList );
}


extern void List_insertionSort ( List_t * const pList )
/*
 TASK : Performs an insertion sort on the list pList. This is the best sort
        to use when the list contains about 50 nodes or less.
*/
{
   assert (pList != NULL);
   List_insertionSortLink ( &(pList->pHeadNode), &(pList->finalNode), pList->pfnCompare );
   List_repairPriorLinks ( pList );
}


extern void List_deduplicate ( List_t * const pList )
/*
 TASK : Takes a sorted list and deduplicates all data entries leaving only unique
        values
 NOTE : The list must be sorted for this function to remove all duplicate values
*/
{
   ListNode_t *pTempNode = NULL;
   long count = 1;

   assert ( pList != NULL );
   if ( pList->pHeadNode == &(pList->finalNode ))
      return;
   for ( pTempNode = pList->pHeadNode; pTempNode->pNextNode != &(pList->finalNode); )
   {
      if (0 == pList->pfnCompare ( pTempNode->pvNodeData, pTempNode->pNextNode->pvNodeData ))
      {
         pList->pCurrentNode = pTempNode->pNextNode;
         pList->numberOfCurrentNode = count + 1;
         List_removeCurrent (pList);
      }
      else
      {
         pTempNode = pTempNode->pNextNode;
         ++count;
      }
   }
   pList->walkingListStatus = currentlyNotWalkingList;
}


/***************************************************************************/
/*                                                                         */
/*                          Printing Operator                              */
/*                                                                         */
/***************************************************************************/


extern void List_printList ( List_t * const pList,
                              FILE * pfOutput,
                               FileAccessFunction_t const * pfnPrintData )
/*
 TASK : The user supplies a function to print the contents of a pList
        node data field in pfnPrintData and this prints the entire list
        to pfOutput
*/
{
   void * pVoid;

   assert ( pList != NULL );
   if ( pList == NULL || pfOutput == NULL )
      return;

   List_moveToHead (pList);

   fprintf ( pfOutput, "\nList dump:\n" );
   while ( NULL != ( pVoid = List_walk ( pList )))
   {
      fprintf ( pfOutput, "\n\nNode (%d):\n", pList->numberOfCurrentNode );
      pfnPrintData ( pfOutput, pList->pCurrentNode->pvNodeData );
   }
   fprintf ( pfOutput, "\n\n" );
   fflush ( pfOutput );
}


static void List_insertionSortLink ( ListNode_t ** listHead,
                                      ListNode_t * listEnd,
                                       CompareFunction_t Compare )
/*
 TASK : Performs an insertion sort on the list beginning at listHead and
        and ending at listEnd.
 NOTE : This only sorts the list by the pNextNode links. The pPriorNode links
        have to be reestablished correctly using List_repairPriorLinks after calls
        to this function
*/
{
   ListNode_t *newlist;
   ListNode_t *walk, *save;

   newlist = listEnd;
   walk = *listHead;

   for ( ; walk != listEnd; walk = save )
   {
      ListNode_t ** pnewlink;
      for ( pnewlink = &newlist;
             *pnewlink != listEnd && Compare ( walk->pvNodeData, (*pnewlink)->pvNodeData) >= 0;
               pnewlink = &((*pnewlink)->pNextNode ));
      save = walk->pNextNode;
      walk->pNextNode = *pnewlink;
      *pnewlink = walk;
   }
   *listHead = newlist;
}


static void List_repairPriorLinks ( List_t * const pList )
/*
 TASK : The sorting routines disrupt the nodes prior pointers. Instead of
        ensuring these are correct during the sorting routines it is more
        economical to reestablish the correct prior pointer order after
        sorting by traversing the forward links.
        This function reestablishes the pPriorNode order of pList
*/
{
   ListNode_t *walk, *save;

   if ( pList->pHeadNode != &(pList->finalNode ))
   {
      walk = pList->pHeadNode->pNextNode;
      save = pList->pHeadNode;
      while ( walk != &(pList->finalNode))
      {
         walk->pPriorNode = save;
         save = walk;
         walk = walk->pNextNode;
      }
      walk->pPriorNode = save;
      pList->finalNode.pNextNode = pList->pHeadNode;
      pList->pHeadNode->pPriorNode = &(pList->finalNode);
      pList->pCurrentNode = pList->pHeadNode;
      pList->numberOfCurrentNode = 1;
   }
}


static void List_xQuickSortL2 ( ListNode_t **Head, ListNode_t * End, long N )
/*
 TASK : Performs a quick sort on the list beginning Head and ending End
 NOTE : Uses an insertion sort if the sublist is <= 7 nodes long
        Eliminates tail recursion to minimise stack size.
        Uses a randomly selected pivot to avoid problems with pre-ordered
        data.
        Adapted from artice in "C Gazette", Vol 5, No. 6 (1995)
*/
{
   long left_count, right_count, npivot;
   ListNode_t **left_walk, *pivot, *old;
   ListNode_t **right_walk, *right;

   while ( N > 1 )
   {
      if ( N <= 7 ) /* insertion sort small sublists */
      {
         List_insertionSortLink ( Head, End, storedCompare );
         break;
      }
      npivot = abs ( rand ()) % N;
      if ( npivot < 2 || npivot > N - 2 )
         npivot = 2;
      old = *Head;
      while ( npivot-- )
         old = old->pNextNode;
      pivot = old->pNextNode;
      old->pNextNode = pivot->pNextNode;

      left_walk = Head;
      right_walk = &right;
      left_count = right_count = 0;
      for ( old = * Head; old != End; old = old->pNextNode )
      {
         if ( storedCompare ( old->pvNodeData, pivot->pvNodeData ) < 0 )
         {
            ++left_count;
            *left_walk = old;
            left_walk = &(old->pNextNode);
         }
         else
         {
            ++right_count;
            *right_walk = old;
            right_walk = &(old->pNextNode);
         }
      }
      *right_walk = End;
      *left_walk = pivot;
      pivot->pNextNode = right;
      if ( left_count > right_count )
      {
         List_xQuickSortL2 (&(pivot->pNextNode), End, right_count );
         End = pivot;
         N = left_count;
      }
      else
      {
         List_xQuickSortL2 ( Head, pivot, left_count );
         Head = &(pivot->pNextNode);
         N = right_count;
      }
   }
}


