/***************************************************************************/
/*                                                                         */
/*                             DLLIST.H                                    */
/*                                                                         */
/*                     Generic doubly linked list                          */
/*                                                                         */
/*                                                                         */
/***************************************************************************/

/*
    Uses a doubly linked circular list because :

      Simplifies insertions/deletions.
      Allows both backward and forward traversal of the list.
      Allows the list to also be used as a stack, ring or queue.
*/



#ifndef _DLLIST_H_INCLUDED

#include <assert.h>
#include <stdlib.h>
#include <stdio.h>


#ifdef __cplusplus
extern "C"
{
#endif

/* Generic comparison function defined as a type                           */
typedef int CompareFunction_t ( const void *, const void * );
/* Generic memory allocation function defined as a type                    */
typedef void * CreateObjectFunction_t ( const void * );
/* Generic memory deallocation function defined as a type                  */
typedef void FreeObjectFunction_t ( void * );
/* Generic file access function defined as a type                          */
typedef void FileAccessFunction_t ( FILE *, void * );



// Create a boolean type

#define BOOL_TRUE  1 == 1
#define BOOL_FALSE !BOOL_TRUE

typedef enum { Bfalse = BOOL_FALSE, Btrue = BOOL_TRUE } Boolean;



#pragma pack ( push, 1 );        /* Align all structures on 1 byte boundaries */

   typedef struct node
   {
      void * pvNodeData;         /* Pointer to the data contained by the node */
      struct node *pNextNode;    /* Pointer to next node in list              */
      struct node *pPriorNode;   /* Pointer to previous node in list          */
   } ListNode_t;                 /* Format of one list node                   */

   typedef enum
   {
      currentlyWalkingList = 0,
      currentlyNotWalkingList = 1
   } WalkStatus;

   typedef struct
   {
      ListNode_t *pHeadNode;     /* Pointer to the head node of the list      */
      ListNode_t *pCurrentNode;  /* Pointer to the current node in the list   */
      ListNode_t finalNode;      /* Permanent dummy node marking list end     */
      long numberOfNodes;        /* The number of nodes in the list 1..n      */
      long numberOfCurrentNode;  /* The number of the current list node 1..n  */
      WalkStatus walkingListStatus;    /* Storage for walk status             */
      /* pointers to user supplied functions                                  */
      CompareFunction_t * pfnCompare;
      CreateObjectFunction_t  * pfnCreateData;
      FreeObjectFunction_t * pfnFreeData;
   } List_t;                     /* Format of a list                          */


#pragma pack ( pop );



/***************************************************************************/
/*                                                                         */
/*                         Constructor/Destructor                          */
/*                                                                         */
/***************************************************************************/


extern List_t * List_new ( CompareFunction_t * pfnCompare,
                            CreateObjectFunction_t * pfnCreateData,
                             FreeObjectFunction_t * pfnFreeData );
/*
 TASK : Creates a new list.
 IN   : The size of the data type this list is to hold - NodeDataSize
 OUT  : Pointer to the newly created list. If NULL is returned, the list
        could not be created.
*/

extern void List_delete( List_t **ppList);
/*
 TASK : Deletes the list.
*/


/***************************************************************************/
/*                                                                         */
/*                            Enquiry Operators                            */
/*                                                                         */
/***************************************************************************/


extern Boolean List_isEmpty (List_t * const pList);
/*
 TASK : Returns Btrue if the list pointed to by pList contains no nodes.
*/

extern long List_currentPosition (List_t * const pList);
/*
 TASK : Returns the position in the list of the current node ( Head node
        ordinal position is 1 )
*/

extern Boolean List_notAtEnd (List_t * const pList);
/*
 TASK : Returns Btrue if the current node is not the tail node.
*/

extern long List_numberOfNodes (List_t * const pList);
/*
 TASK : Returns the number of nodes in the list.
*/



/***************************************************************************/
/*                                                                         */
/*                         Positional Operators                            */
/*                                                                         */
/***************************************************************************/



extern Boolean List_moveToHead (List_t * const pList);
/*
 TASK : Makes the head node the current node. If the list is empty, returns
        Bfalse, otherwise returns Btrue.
 NOTE : Incompatable with sequential traversal.
*/

extern Boolean List_moveToTail (List_t * const pList);
/*
 TASK : Makes the tail node the current node
 NOTE : Incompatable with sequential traversal.
*/

extern Boolean List_moveNthNode (List_t * const pList, long MoveToThisNode);
/*
 TASK : Makes the nth node of the list the current node. Returns Bfalse if the
        list is empty, or MoveToThisNode is greater than the number of nodes
        in the list, otherwise returns Btrue.
 NOTE : Incompatable with sequential traversal. The head node position is 1.
*/

extern Boolean List_moveToNextNode (List_t * const pList);
/*
 TASK : Makes the next node of the list the current node. If the tail of
        the list is reached moves to the list head and returns Bfalse,
        otherwise returns Btrue.
 NOTE : Incompatable with sequential traversal.
*/

extern Boolean List_moveToPreviousNode (List_t * const pList);
/*
 TASK : Makes the previous node in the list the current node. If the head of
        the list the list is reached moves to the list tail and returns Bfalse
        otherwise returns Btrue.
 NOTE : Incompatable with sequential traversal.
*/


extern Boolean List_makeCurrentHead ( List_t * const pList );
/*
 TASK : Makes the current node the head of the list.
        If the list is empty then returns Bfalse, otherwise returns Btrue
*/



/***************************************************************************/
/*                                                                         */
/*                         Extraction Operators                            */
/*                                                                         */
/***************************************************************************/



extern void * List_data ( List_t * const pList );
/*
 TASK : Returns a pointer to the contents of the current node.
*/


extern void * const List_next ( List_t * const pList );
/*
 TASK : Returns a pointer to the next node in the list or NULL if there
        is no next node
*/



extern ListNode_t * List_nodePointer ( List_t * const pList );
/*
 TASK : Returns a pointer to the current node.
*/



extern void * const List_walk ( List_t * const pList );
/*
 TASK : Sequentially returns a pointer to the contents of each node in the
        list starting at the head node and moving to the tail.
        Returns NULL if the list is empty or if reached the end of list.
        Moves to head node when traversal complete.

        Always call List_moveToHead before List_walk if you need to be
        absolutely certain the list walk will start from the first node

 NOTE : IT IS CRITICALLY IMPORTANT THAT IF YOU TERMINATE WALKING THE LIST
        BEFORE THE NULL POINTER IS RETURNED YOU CALL List_moveToHead TO
        RESET THE LIST STATE.
*/


extern void * const List_walkFromNth ( List_t * const pList, long NodeNumber );
/*
 TASK : Sequentially returns a pointer to the contents of each node in the
        list starting at the NodeNumber node and moving to the tail.
        Returns NULL if the list is empty or if reached the end of list,
        or if NodeNumber is > number of nodes in the list
        Moves to head node when traversal complete.

        Always call List_moveToHead before List_walkFromNth if you need
        to be absolutely certain the walk will start from the first node

 NOTE : IT IS CRITICALLY IMPORTANT THAT IF YOU TERMINATE WALKING THE LIST
        BEFORE THE NULL POINTER IS RETURNED YOU CALL List_moveToHead TO
        RESET THE LIST STATE.
*/



extern void * const List_walkAndFind ( List_t * const pList, const void * pvData );
/*
 TASK : Starting at the head node and traversing towards the tail of the list
        this function returns a pointer to the next instance of a node found
        to be identical to the data referenced by pvData according to the
        comparison function pfnCompare. Returns NULL if no more hits are found
        or the list is empty.

        Always call List_moveToHead before List_walkAndFind if you need
        to be absolutely certain the search will start from the first node


 NOTE : IT IS CRITICALLY IMPORTANT THAT IF YOU TERMINATE WALKING THE LIST
        BEFORE THE NULL POINTER IS RETURNED YOU CALL List_moveToHead TO
        RESET THE LIST STATE.
*/


extern void * const List_firstMatch ( List_t * const pList, const void * pvData );
/*
 TASK : Assuming the list is walked from the first node to the last, this
        will return a pointer to the first node whose data is found to
        match pvData and also make that node current. Returns NULL if
        no match was found.
*/



/***************************************************************************/
/*                                                                         */
/*                          Inclusion Operators                            */
/*                                                                         */
/***************************************************************************/


extern void List_replaceCurrent ( List_t * const pList, const void * pvData );
/*
 TASK : Replaces the contents of the current node with a copy of the data
        referenced via pvData
*/

extern Boolean List_addAfter ( List_t * const pList, const void * pvData);
/*
 TASK : Adds a new node to the list containing a copy of the data referenced
        by pvData after the current node. Returns Btrue if memory was
        avaliable to create the new node, otherwise returns Bfalse.
*/

extern Boolean List_addBefore ( List_t * const pList, const void * pvData );
/*
 TASK : Adds a new node to the list containing a copy of the data referenced
        by pvData before the current node. Returns Btrue if memory was
        avaliable to create the new node, otherwise returns Bfalse.
*/

extern Boolean List_addList ( List_t * const pListA, List_t * const pListB);
/*
 TASK : Adds the contents of listB after the current node of listA.
        Returns Btrue if memory was avaliable for the operation,
        otherwise returns Bfalse.
*/


extern Boolean List_appendList  ( List_t * const pListA, List_t * const pListB );
/*
 TASK : Appends a copy of pListB from the current node in ListB to the
        end of ListB to the end of pListA. The list referenced
        by pListB is unaffected by the append. Returns Btrue if memory was
        avaliable to make the copy of pListB, otherwise returns Bfalse.
*/



/***************************************************************************/
/*                                                                         */
/*                           Deletion Operators                            */
/*                                                                         */
/***************************************************************************/



extern Boolean List_removeCurrent (List_t * const pList);
/*
 TASK : Deletes the current node from the list. Returns Bfalse if the list is
        empty, otherwise returns Btrue.
*/


extern void List_empty ( List_t * const pList );
/*
 TASK : Removes all the nodes in the list pointed to by pList
*/


/***************************************************************************/
/*                                                                         */
/*                           Sorting Operators                             */
/*                                                                         */
/***************************************************************************/


extern void List_quickSort ( List_t * const pList );
/*
 TASK : Performs a sort on pList using the quicksort algorithm.
 NOTE : Eliminates tail recursion to minimise stack size.
        Uses a randomly selected pivot to avoid problems with pre-ordered
        data.
        Adapted from artice in "C Gazette", Vol 5, No. 6 1995
*/


extern void List_insertionSort ( List_t * const pList );
/*
 TASK : Performs an insertion sort on the list pList. This is the best sort
        to use when the list contains about 50 nodes or less.
*/


extern void List_deduplicate ( List_t * const pList );
/*
 TASK : Takes a sorted list and deduplicates all data entries leaving only unique
        values
 NOTE : The list must be sorted for this function to remove all duplicate values
*/



/***************************************************************************/
/*                                                                         */
/*                          Printing Operator                              */
/*                                                                         */
/***************************************************************************/


extern void List_printList ( List_t * const pList,
                              FILE * pfOutput,
                               FileAccessFunction_t const * pfnPrintData );
/*
 TASK : The user supplies a function to print the contents of a pList
        node data field in pfnPrintData and this prints the entire list
        to pfOutput
*/



#define _DLLIST_H_INCLUDED
#ifdef __cplusplus
};
#endif

#endif


